
(load "sort.scm")
(load "associators.scm")
(load "lalr.scm")

(if (defined? '%package-data-dir)
   (debug-set! stack 200000)) ; set guile's stack checking higher than normal

(display ";;; building grammar...") (newline)

;;;;;
;;
;; ripped off from example in lalr (Mark Johnson (mj@cs.brown.edu))
;;
;;;;;


(define (identity expr) expr)
(define (unary-apply op expr1) (list op expr1))
(define (binary-apply expr1 op expr2) (list op expr1 expr2))
(define (quant-apply op expr1 col expr2 col1 expr3)
  (list op expr1 expr2 expr3))


(define (string-match-pos s m p)
  (if (and (>= (string-length s) (+ (string-length m) p))
	   (string=? m (substring s p (+ p (string-length m)))))
      #t #f))

;; from the little schemer
(define member?
  (lambda (a lat)
    (cond
     ((null? lat) #f)
     (else (or (eq? (car lat) a)
	       (member? a (cdr lat)))))))


(define expr-grammar
  `(;; change ` to ' if you want to generate a Scheme program table 

    ;; : used as a variable and range pair  f(i:0<i<5)
    ;;(termc --> term1 term-col term1 ,binary-apply)
    ;;(termc --> term1 ,identity)

    ;; <==>  <=!=>
    (term1 --> term1 term-op1 term2 ,binary-apply)
    (term1 --> term2 ,identity)

    ;; ==> <==
    (term2 --> term2 term-op2 term3 ,binary-apply)
    (term2 --> term3 ,identity)

    ;; \/  /\
    (term3 --> term3 term-op3 term4 ,binary-apply)
    (term3 --> term4 ,identity)

    ;; =  <  >  <=  >=  !=
    (term4 --> term4 term-op4 term5 ,binary-apply)
    (term4 --> term5 ,identity)

    ;; +
    (term5 --> term5 term-op5 term6 ,binary-apply)
    ;;(term5 --> term5 term-op-neg term6 ,binary-apply) ; use x + (-y)
    (term5 --> term6 ,identity)

    ;; min  max  *  / div
    (term6 --> term6 term-op6 term7 ,binary-apply)
    (term6 --> term7 ,identity)

    ;; - ~  ,
    (term7 --> term7 term-op7 term-leaf ,binary-apply)
    (term7 --> term-op-bneg term7 ,unary-apply)
    (term7 --> term-op-neg term-leaf ,unary-apply)
    (term7 --> term-func ,identity)

    ;; function application
    ;; f.x
    (term-func --> term-func term-op-dot term-leaf ,binary-apply)
    ;; f(x)
    (term-func --> var lparen term1 rparen
	   ,(lambda (expr1 paren1 expr2 paren2)
	      (binary-apply expr1 'dot expr2)))
    (term-func --> term-leaf ,identity)

    ;; quantified expression
    (term-func --> term-op-quant term1 term-col term1 term-col term1
	       ,quant-apply)
    ;; empty range
    (term-func --> term-op-quant term1 term-col term-col term1
	       ,(lambda (op expr1 col col1 expr3)
		  (quant-apply op expr1 col #t col1 expr3)))
    ;; empty term
    (term-func --> term-op-quant term1 term-col term1 term-col
	       ,(lambda (op expr1 col expr2 col1)
		  (quant-apply op expr1 col expr2 col1 #f)))
    ;; empty range and term
    (term-func --> term-op-quant term1 term-col term-col
	       ,(lambda (op expr1 col col1)
		  (quant-apply op expr1 col #t col1 #f)))

    (term-leaf --> lparen term1 rparen ,(lambda (lp term rp) term))
    (term-leaf --> integer ,identity)
    (term-leaf --> var ,identity)
    (term-leaf --> infinity ,identity)
    (term-leaf --> constant ,identity)
    (term-leaf --> term-func ,identity)

    (var --> chr var ,(lambda (c v) (string-append (make-string 1 c) v)))
    (var --> chr ,(lambda (c) (make-string 1 c)))

    (integer --> term-op-neg number ,(lambda (neg i) (- i)))
    (integer --> number ,identity)

    (number --> number digit ,(lambda (n d) (+ (* 10 n) d)))
    (number --> digit ,identity)
    ))


; (define expr-grammar
;   `((expr --> expr expr-op term ,binary-apply)    ;;; change ` to '
;     (expr --> term ,identity)                     ;;; if you want to generate
;     (term --> term term-op term1 ,binary-apply)   ;;; a Scheme program table
;     (term --> term1 ,identity)
;     (term1 --> lparen expr rparen ,(lambda (lp expr rp) expr))
;     (term1 --> number ,identity)
;     (number --> number digit ,(lambda (n d) (+ (* 10 n) d)))
;     (number --> digit ,identity)))


(define expr-terminals '(term-op1
 			 term-op2
 			 term-op3
 			 term-op4
 			 term-op5
 			 term-op6
			 term-op7
 			 term-op-neg
 			 term-op-bneg ; boolean negate
 			 term-col
 			 term-op-dot
 			 term-op-quant
 			 lparen
 			 rparen
 			 digit
 			 chr
 			 constant))

;(define expr-terminals '(expr-op term-op lparen rparen digit))

(define table (lalr-table expr-grammar expr-terminals #f))

(define (parse-gcl string)
  (let ((pos 0)
	(strings-length (string-length string)))

    (define (lexical-analyser)
      (if (< pos strings-length)
	  (let ((char0 (string-ref string pos))
		(char1 (if (< (+ pos 1) strings-length)
			   (string-ref string (+ pos 1))
			   #\ )))

	    ;; binding power is encoded here by which term-op
	    ;; the token belongs to.  term-op6 binds more tightly
	    ;; than term-op5.  the ranking can be divined from
	    ;; the definition of expr-grammar.

	    (if (char-whitespace? char0)
		(begin
		  (set! pos (+ pos 1))
		  (lexical-analyser))
		(cond
		 ((string-match-pos string "true" pos)
		  (set! pos (+ pos 4)) `(constant . ,#t))
		 ((string-match-pos string "false" pos)
		  (set! pos (+ pos 5)) `(constant . ,#f))
		 ((string-match-pos string "inf" pos)
		  (set! pos (+ pos 3)) `(constant . ,'inf))
		 ((string-match-pos string "AND" pos)
		  (set! pos (+ pos 3)) `(term-op-quant . ,'AND))
		 ((string-match-pos string "OR" pos)
		  (set! pos (+ pos 2)) `(term-op-quant . ,'OR))
		 ((string-match-pos string "MAX" pos)
		  (set! pos (+ pos 3)) `(term-op-quant . ,'MAX))
		 ((string-match-pos string "MIN" pos)
		  (set! pos (+ pos 3)) `(term-op-quant . ,'MIN))
		 ((string-match-pos string "SUM" pos)
		  (set! pos (+ pos 3)) `(term-op-quant . ,'SUM))
		 ((string-match-pos string "PROD" pos)
		  (set! pos (+ pos 4)) `(term-op-quant . ,'PROD))
		 ((string-match-pos string "N" pos)
		  (set! pos (+ pos 1)) `(term-op-quant . ,'N))
		 ((string-match-pos string "max" pos)
		  (set! pos (+ pos 3)) `(term-op6 . ,'max))
		 ((string-match-pos string "min" pos)
		  (set! pos (+ pos 3)) `(term-op6 . ,'min))
		 ;;((string-match-pos string "and" pos)
		 ;; (set! pos (+ pos 3)) `(term-op1 . ,'and))
		 ;;
		 ((string-match-pos string "<=!=>" pos)       ; <=!=>
		  (set! pos (+ pos 5)) `(term-op1 . ,'<=!=>))
		 ((string-match-pos string "<==>" pos)        ; <==>
		  (set! pos (+ pos 4)) `(term-op1 . ,'<==>))
		 ((string-match-pos string "==>" pos)         ; ==>
		  (set! pos (+ pos 3)) `(term-op2 . ,'==>))
		 ((string-match-pos string "<==" pos)         ; <==
		  (set! pos (+ pos 3)) `(term-op2 . ,'<==))
		 ((and (eqv? char0 #\>) (eqv? char1 #\=))     ; >=
		  (set! pos (+ pos 2)) `(term-op4 . ,'>=))
		 ((and (eqv? char0 #\<) (eqv? char1 #\=))     ; <=
		  (set! pos (+ pos 2)) `(term-op4 . ,'<=))
		 ((and (eqv? char0 #\!) (eqv? char1 #\=))     ; !=
		  (set! pos (+ pos 2)) `(term-op4 . ,'!=))
		 ;;
		 ((and (eqv? char0 #\\) (eqv? char1 #\/))     ; \/
		  (set! pos (+ pos 2)) `(term-op3 . ,'\/))
		 ((and (eqv? char0 #\/) (eqv? char1 #\\))     ; /\
		  (set! pos (+ pos 2)) `(term-op3 . ,'/\))
		 ;;
		 ((eqv? char0 #\+) (set! pos (+ pos 1)) `(term-op5 . ,'+))
		 ((eqv? char0 #\-) (set! pos (+ pos 1)) `(term-op-neg . ,'-))
		 ((eqv? char0 #\*) (set! pos (+ pos 1)) `(term-op6 . ,'*))
		 ((eqv? char0 #\/) (set! pos (+ pos 1)) `(term-op6 . ,'/))
		 ((string-match-pos string "div" pos)         ; div
		  (set! pos (+ pos 3)) `(term-op6 . ,'div))

		 ((eqv? char0 #\~) (set! pos (+ pos 1)) `(term-op-bneg . ,'~))
		 ((eqv? char0 #\!) (set! pos (+ pos 1)) `(term-op-bneg . ,'~))
		 ((eqv? char0 #\=) (set! pos (+ pos 1)) `(term-op4 . ,'=))
		 ((eqv? char0 #\>) (set! pos (+ pos 1)) `(term-op4 . ,'>))
		 ((eqv? char0 #\<) (set! pos (+ pos 1)) `(term-op4 . ,'<))
		 ((eqv? char0 #\,) (set! pos (+ pos 1)) `(term-op7 . ,'comma))
		 ((eqv? char0 #\.) (set! pos (+ pos 1)) `(term-op-dot . ,'dot))
		 ((eqv? char0 #\:) (set! pos (+ pos 1)) `(term-col . ,'col))
		 ;;
		 ((eqv? char0 #\() (set! pos (+ pos 1)) '(lparen . #f))
		 ((eqv? char0 #\)) (set! pos (+ pos 1)) '(rparen . #f))
		 ;;
		 ((member? char0 (list #\0 #\1 #\2 #\3 #\4
				       #\5 #\6 #\7 #\8 #\9))
		  (set! pos (+ pos 1))
		  `(digit . ,(- (char->integer char0) (char->integer #\0))))
		 ((member? char0 (list #\a #\b #\c #\d #\e #\f
				       #\g #\h #\i #\j #\k #\l
				       #\m #\n #\o #\p #\q #\r
				       #\s #\t #\u #\v #\w #\x
				       #\y #\z #\_))
		  (set! pos (+ pos 1))
		  `(chr . ,char0))
		 (else (parse-error)))))))

    (define (parse-error)
      (display "Error somewhere in ")
      (write (substring string 0 pos))
      (newline))

    ;; the extra list is here so this returns a list of
    ;;  parsed expressions rather than just one.  this will
    ;;  always return a list of length one, but...
    (list
     (lalr-parser table lexical-analyser parse-error))))
