
#include <stdio.h>
#include <string.h>

#include <gdk/gdk.h>
#include <gtk/gtkwidget.h>
#include <glade/glade.h>
#include <glade/glade-xml.h>

#include <libguile.h>
#include <guile/gh.h>
/* void scm_gtk_init (); */
#include "guile-gtk.h"
#include "scm-yacc-glue.h"
#include "equation-yaccer.h"

#define GTK_CAN_SHOW_8_BIT_LABELS

void init_my_guile_gtk_defs ();

typedef long int (*scmf)();

void sgtk_init_gtk_gtk_glue ();
void sgtk_init_gtk_gdk_glue ();
void init_glade_defs ();


SCM make_symbol_from_string (char *str)
{
# ifndef SCM_MAJOR_VERSION /* then it's guile 1.4 or earlier */
  return gh_symbol2scm (str);
# else
  return scm_str2symbol (str);
# endif
}



GladeXML *gxml;


static char *escape_backslashes (char *in)
{
  char *escaped, *ep;
  int length = strlen (in);

  escaped = (char *) malloc (length * 2 + 5);
  ep = escaped;

  while (*in)
    {
      if ((*in) == '\\')
	{
	  *ep++ = '\\';
	  *ep++ = '\\';
	}
      else
	*ep++ = *in;
      in++;
    }

  *ep = '\0';

  return escaped;
}


void main_prog (void *closure, int argc, char **argv)
{
  char *scratch;
  char *escaped_argv1;
  int i, arg_lengths = 0;

  SGTK_REGISTER_GLUE (sgtk_init_gtk_gtk_glue);
  SGTK_REGISTER_GLUE (sgtk_init_gtk_gdk_glue);
  init_glade_defs ();

  gtk_init (&argc, &argv);
  glade_init ();

  sgtk_init (argc, argv);
  init_my_guile_gtk_defs ();

  if (argc > 1)
    escaped_argv1 = escape_backslashes (argv[ 1 ]);
  else
    escaped_argv1 = strdup ("");

  for (i=1; i < argc; i++)
    arg_lengths += strlen (argv[ i ]);

  /* count argv[ 1 ] twice because it gets backslash escaped
     and could end up being twice as long as it started */
  scratch = (char *) malloc (60 + arg_lengths +
			     (argc > 1 ? strlen (argv[ 1 ]) : 0) +
			     argc*3);

  sprintf (scratch,
	   "(load \"gcl-rewriter.scm\")\n"
	   "(main (list \"gcl-rewriter\" \"%s\"",
	   escaped_argv1);

  for (i=2; i<argc; i++)
    {
      strcat (scratch, " \"");
      strcat (scratch, argv[ i ]);
      strcat (scratch, "\"");
    }

  strcat (scratch, "))\n");

# if 1
  gh_eval_str_with_stack_saving_handler (scratch);
# else
  /*
  printf ("(load \"gcl-rewriter.scm\")\n"
	  "(main (list \"gcl-rewriter\" \"%s\"))\n",
	  escaped_argv1);
  */
  printf ("%s", scratch);
  scm_shell (0, NULL);
  /*sgtk_shell (argc, argv);*/
# endif

  free (escaped_argv1);
  free (scratch);
}


int main (int argc, char *argv[])
{
  g_log_set_always_fatal ((GLogLevelFlags) 0xFFFF);

  /* gh_enter (argc, argv, main_prog); */

  /* printf ("argv1 = \"%s\"\n", argc>1 ? argv[ 1 ] : "missing"); */

  scm_boot_guile (argc, argv, main_prog, 0);
  return 0; /* unreached */
}

/**********************************************************************/


SCM s_parse_gcl (SCM scm_expressions)
{
  char *expressions = SCM_CHARS (scm_expressions);
  return parse_gcl (expressions);
}


SCM s_gtk_widget_set_font_to_symbol (SCM p_widget)
{
  static GtkStyle *symbol_style = NULL;
  GtkWidget *c_widget;

  SCM_ASSERT (sgtk_is_a_gtkobj (gtk_widget_get_type (), p_widget),
	      p_widget, SCM_ARG1, "s_gtk_label_set_font_to_symbol");

  SCM_DEFER_INTS;

  c_widget = (GtkWidget *) sgtk_get_gtkobj (p_widget);

  if (symbol_style == NULL)
    {
      char symbol_font_name[ 300 ];
      GdkFont *symbol;
      GtkStyle *default_style = gtk_widget_get_style (c_widget);

#warning "this used to work, but it seems that gtk now shows a blank label if any high bits are set...?"

      /* -adobe-symbol-medium-r-normal--12-120-75-75-p-74-adobe-fontspecific */

      snprintf (symbol_font_name, sizeof (symbol_font_name),
		"-adobe-symbol-*-*-normal-*-%d-*-*-*-*-*-*-*",
		default_style->font->ascent
		+ 1 /* hatred. */
		/*+ default_style->font->descent*/);

      symbol = gdk_font_load (symbol_font_name);

      /*printf ("s_gtk_widget_set_font_to_symbol: %d, %d\n",
	      default_style->font->ascent,
	      symbol->ascent);*/

      symbol_style = gtk_style_copy (default_style);
      symbol_style->font = symbol;
    }

  gtk_widget_set_style (c_widget, symbol_style);

  SCM_ALLOW_INTS;

  return SCM_MAKINUM (SCM_BOOL_T);
}



void init_my_guile_gtk_defs ()
{
# ifdef GTK_CAN_SHOW_8_BIT_LABELS
  scm_make_gsubr ("gtk-widget-set-font-to-symbol", 1, 0, 0,
		  (scmf) s_gtk_widget_set_font_to_symbol);
# endif
  scm_make_gsubr ("parse-gcl", 1, 0, 0, (scmf) s_parse_gcl);
}


/*************************************************************************/
/*************************************************************************/
/*************************************************************************/

SCM new_number_exp (int i)
{
  return SCM_MAKINUM (i);
}


SCM new_boolean_exp (int b)
{
  return b ? SCM_BOOL_T : SCM_BOOL_F;
}


SCM new_variable_exp (char *vn)
{
  return scm_makfrom0str (vn);
}


SCM new_quant_operator (char *operator_string)
{
  return make_symbol_from_string (operator_string);
}


SCM new_infix_exp (char *operator_string, SCM left, SCM right)
{
  /* keep this synced with gcl-symbol->string in gcl.scm */

  if (strcmp (operator_string, ",") == 0)
    operator_string = "comma";   /* XXX strdup these? */
  else if (strcmp (operator_string, ":") == 0)
    operator_string = "col";
  else if (strcmp (operator_string, ".") == 0)
    operator_string = "dot";

  return scm_cons (make_symbol_from_string (operator_string),
         scm_cons (left,
         scm_cons (right,
		   SCM_EOL)));
}


SCM new_unary_exp (char *operator_string, SCM a)
{
  return scm_cons (make_symbol_from_string (operator_string),
         scm_cons (a,
		   SCM_EOL));
}


SCM new_quant_exp (SCM operator, SCM dummies, SCM range, SCM term)
{
  return scm_cons (operator,
         scm_cons (dummies,
         scm_cons (range,
         scm_cons (term,
		   SCM_EOL))));
}


SCM new_type_exp (char *type_name)
{
  return scm_makfrom0str (type_name);
}


SCM new_exp_list ()
{
  return SCM_EOL;
}


SCM append_to_exp_list (SCM el, SCM e)
{
  return scm_cons (e, el);
}
