#!/usr/local/bin/guile \
--debug -e main -s
!#

(use-modules (gtk gtk) (gtk gdk))

(debug-enable 'backtrace 'debug)

(define rules-files-directory "./rules/")

(define rules-files '(("Equivalence" ("true-is-unit"
				      "symmetric"
				      "associative"
				      "discrepancy-is-symmetric"
				      "discrepancy-associates"
				      "discrepancy-with-negation"))
		      ("Disjunction" ("associative"
				      "symmetric"
				      "distributes-over-equivalence"
				      "distributes-over-disjunction"
				      "distributes-over-conjunction"
				      "absorption"
				      "true-is-zero"
				      "false-is-unit"
				      "idempotent"))
		      ("Golden"      ("part-a"
				      "part-b"
				      "part-c"))
		      ("Conjunction" ("symmetric"
				      "associative"
				      "idempotent"
				      "distributes-over-conjunction"
				      "distributes-over-even-equivalence"
				      "distributes-over-discrepancy"
				      "distributes-over-disjunction"
				      "absorption"
				      "true-is-unit"
				      "false-is-zero"))
		      ("Implication" ("definition"
				      "x-implies-x"
				      "x-and-y-implies-x"
				      "x-implies-x-or-y"
				      "true-implies-x"
				      "with-conjunction"
				      "x-implies-y-implies-z"
				      "x-and-x-implies-y"
				      "transitive"
				      "x-implies-y-or-y-implies-z"
				      "consequence-to-implication"))
		      ("Negation"    ("definition"
				      "switch"
				      "twice"
				      "to-equivalence"
				      "De-Morgan"
				      "complement"
				      "with-implication"
				      "contrapositive"
				      "shuffle"
				      "true-false"
				      "x-is-not-x"
				      "false-implies-anything"
				      "contradiction"))
		      ("Arithmetic"  ("double-negative"
				      "negative-distributes-over-addition"
				      "addition-is-symmetric"
				      "addition-is-associative"
				      "multiplication-is-symmetric"
				      "multiplication-is-associative"
				      "multiplication-distributes-over-add"
				      "at-least-to-at-most"
				      "unit-and-zero-of-multiplication"
				      "unit-and-zero-of-addition"))
		      ("MaxMinAbs"   ("definition-of-abs"
				      "associative"
				      "symmetric"
				      "idempotent"
				      "addition-distributes"
				      "unit-and-zero"
				      "at-most-at-least"
				      "with-negative"))
		      ("Quantifiers" ("term-rule"
				      "range-rule"
				      "distributivity"
				      "empty-range"
				      "constant-term"
				      "trading"
				      "de-morgan"
				      "one-point"
				      "arithmetic"))))

(load "gcl.scm")

(define (gtk-widget-show-multi . widgets)
  (for-each gtk-widget-show widgets))


;; from Marius...

(define (gtk-signal-connect* widget signal callback)
   (gtk-signal-connect widget signal callback))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



(define (create-math-label r)

  (if (defined? 'gtk-widget-set-font-to-symbol)
      ;; make an hbox of labels, switching fonts as needed.
      ;;  this is so we can show the correct math symbols for some things.
      (let ((hbox (gtk-hbox-new #f 0)))

	(define (output-segment str in-symbol-font)
	  (if (not (null? str))
	      (let ((label (gtk-label-new (list->string (reverse str)))))
		(if in-symbol-font (gtk-widget-set-font-to-symbol label))
		(gtk-misc-set-alignment label 0.0 0.0)
		(gtk-misc-set-padding label 0.0 0.0)
		(gtk-box-pack-start hbox label #f #f 0)
		#t)
	      #t))

	(define (cdrn lat n)
	  (if (= n 0)
	      lat
	      (cdrn (cdr lat) (- n 1))))

	(define (consume-symbols a b)
	  (if (null? b)
	      (output-segment a #t)
	      (let ((sym-info (convert-to-symbol-font b)))
		(if (car sym-info)
		    ;; this can be in symbol font, try for more.
		    (consume-symbols (cons (cadddr sym-info) a)
				     (cdrn b (caddr sym-info)))
		    ;; else we have to switch to normal font.
		    (begin
		      (output-segment a #t)
		      (consume-normals '() b))))))

	(define (consume-normals a b)
	  (if (null? b)
	      (output-segment a #f)
	      (let ((sym-info (convert-to-symbol-font b)))
		(if (cadr sym-info)
		    ;; this can be in normal font, try for more.
		    (consume-normals (cons (cadddr sym-info) a)
				     (cdrn b (caddr sym-info)))
		    ;; else we have to switch to normal font.
		    (begin
		      (output-segment a #f)
		      (consume-symbols '() b))))))

	(consume-symbols '() (string->list r))
	hbox)

      ;; else we can't set the widget to use symbol font
      (let ((label (gtk-label-new r)))
	(gtk-misc-set-alignment label 0.0 0.0)
	label)))


;; takes an expression from the command line, and an optional
;;  file name.  the file can contain "givens", a set of rules
;;  which are true for this problem only.
(define (make-gcl-rewriter xml cl-expression givens-files)

  (display "cl-expression: ") (display cl-expression) (newline)
  (display "givens-files: ") (display givens-files) (newline)

  (let ((rewriter-list (glade-xml-get-widget xml "proof-list"))
	(entry (glade-xml-get-widget xml "entry"))
	(rabbits-menu-item (glade-xml-get-widget xml "rabbits-menu-item"))
	(rule-filters '()) ; filter funcs for all the rule lists
	;; used to know the character positions of the contents of "entry"
	(start-indexes '())
	(end-indexes '())
	;; the parsed contents of "entry"
	(current-expression '())
	;; the parsed contents of the selected area of "entry"
	(selected-subtree '())
	(previous-expressions '())
	(previous-rules '()))

    (define (form-rule-group rule-group-spec)
      ;; make a hbox which contains option menus.
      ;; a rule-group-spec looks like this:
      ;; ("Equivalence" ("true-is-unit" "associative-symmetrical" ...))
      (let* ((rule-scrolled-win (gtk-scrolled-window-new))
	     (rule-list-vbox (gtk-vbox-new #f 0)))

	;; make an option-menu with rules from a file
	(define (form-popup rule-file-name)
	  (let* ((file-name 
		  (if (string=? (car rule-group-spec) "Givens")
		      rule-file-name
		      (string-append rules-files-directory
				     (car rule-group-spec) "-"
				     rule-file-name)))
		 (hbox (gtk-hbox-new #f 0))
		 (any-can-button (gtk-check-button-new))
		 (rule-list-option-menu (gtk-option-menu-new))
		 (rule-list-menu (gtk-menu-new))
		 ;; list of parsed rules
		 (rules (load-rules file-name))
		 ;; list items corresponding to rules
		 (rule-list-items '()))

	    ;; add a rule to the end of the list of rules
	    (define (add-rule r)
	      (let* ((r-string (unparse r #f))
		     (label (create-math-label r-string))
		     (item (gtk-menu-item-new)))
		(gtk-container-add item label)
		(gtk-widget-show-all item)
		;; blah, order matters.  could reverse below, but...
		;;(set! rule-list-items (cons item rule-list-items))
		(set! rule-list-items (append rule-list-items (list item)))
		(gtk-menu-append rule-list-menu item)
		(gtk-signal-connect* item "activate"
				     ;; this is so the option menu will
				     ;;  continue to show the top menu item
				     ;;  rather than what was just selected.
				     (lambda ()
				       (gtk-option-menu-set-history
					rule-list-option-menu 0)
				       (applier r)))))

	    ;; ghost out unusable rules (those that can't rewrite "tree")
	    (define (rule-filter tree)
	      (let ((any-can-apply #f))

		(define (rule-filter-rec list-items rule-expressions)
		  (cond
		   ((null? list-items) #t)
		   ((null? rule-expressions) #t)
		   (else
		    (let* ((item (car list-items))
			   (rule (car rule-expressions))
			   (rewritten-and-bindings (apply-rule tree rule)))
		      (cond
		       ((and (atom? (cadr rule))
			     (not (gtk-check-menu-item-active
				   rabbits-menu-item)))
			;; if left of <==> is just a variable then this is
			;;   a rabbit -- ghost it out (unless rabbits are
			;;   enabled.)
			(gtk-widget-set-sensitive item #f))
		       ((eq? 'fail (car rewritten-and-bindings))
			;; this rule doesn't map onto tree, so
			;;  it can't rewrite.
			(gtk-widget-set-sensitive item #f))
		       ((member-equal?
			 (apply-rule-to-subtree current-expression tree rule)
			 previous-expressions)
			;; don't let the proof regress to a previous line.
			;;   the user should use the back button to go back.
			(gtk-widget-set-sensitive item #f))
		       ((not
			 (z-is-dummy-free? rule (cadr rewritten-and-bindings)))
			;; this rule matches, but it does so with
			;;  z matching some part of a quantified expression
			;;  which contains dummie variables.
			(gtk-widget-set-sensitive item #f))
		       (else
			;; this rule can be applied.
			(set! any-can-apply #t)
			(gtk-widget-set-sensitive item #t)))
		      (rule-filter-rec (cdr list-items)
				       (cdr rule-expressions))))))
		(rule-filter-rec rule-list-items rules)
		(gtk-toggle-button-set-state any-can-button any-can-apply)))

	    ;;
	    ;;
	    (gtk-option-menu-set-menu rule-list-option-menu rule-list-menu)
	    (gtk-box-pack-start hbox any-can-button #f #f 0)
	    (gtk-widget-set-sensitive any-can-button #f)
	    (gtk-box-pack-start hbox rule-list-option-menu #t #t 0)
	    (gtk-box-pack-start rule-list-vbox hbox #f #f 0)
	    ;; put a title on the option-menu
	    (gtk-menu-append rule-list-menu
			     (gtk-menu-item-new-with-label rule-file-name))
	    (for-each add-rule rules)
	    (gtk-option-menu-set-history rule-list-option-menu 0)
	    (set! rule-filters (append rule-filters (list rule-filter)))))


	(gtk-scrolled-window-set-policy rule-scrolled-win 'automatic 'always)
	(gtk-widget-set-usize rule-scrolled-win 300 240)
	(gtk-scrolled-window-add-with-viewport rule-scrolled-win
					       rule-list-vbox)

	(for-each form-popup (cadr rule-group-spec))
	(gtk-widget-show-all rule-scrolled-win)

	rule-scrolled-win))
    ;;;;;;;;


    (define (create-notebook rule-group-spec givens-files)
      (let ((notebook (gtk-notebook-new))
	    (rules-and-givens
	     (if (null? givens-files)
		 rule-group-spec
		 (cons (list "Givens" givens-files) rule-group-spec))))
	(gtk-notebook-set-tab-pos notebook 'left)
	(gtk-widget-show notebook)

	(do ((files-loop rules-and-givens (cdr files-loop)))
	    ((null? files-loop))
	  (let* ((tab-name (caar files-loop))
		 (frame (gtk-frame-new tab-name))
		 (label2 (gtk-label-new tab-name))
		 (rule-list (form-rule-group (car files-loop))))
	    (gtk-container-border-width frame 10)
	    (gtk-widget-set-usize frame 200 150)
	    (gtk-widget-show frame)
	    (gtk-widget-show label2)
	    (gtk-container-add frame rule-list)
	    (gtk-notebook-append-page notebook frame label2)))

	`(("next" ,(lambda ()
		     (gtk-notebook-next-page notebook)))
	  ("prev" ,(lambda ()
		     (gtk-notebook-prev-page notebook)))
	  ("rotate" ,(lambda ()
		       (gtk-notebook-set-tab-pos 
			notebook
			(cdr (assq (gtk-notebook-tab-pos notebook)
				   '((top . left)
				     (left . bottom)
				     (bottom . right)
				     (right . top))))))))
	notebook))


    ;; called whenever the users fiddles with "entry".
    ;;  this figures out how much of the expression is selected
    (define (entry-event . e)
      (let ((p (gtk-editable-get-position entry)))
	(let* ((ex (index->expression current-expression
				      (- p 1)       ; the entry widget
				      start-indexes ;   starts counting at 1
				      end-indexes))
	       (start-i (lookup-index ex start-indexes))
	       (end-i (+ (lookup-index ex end-indexes) start-i)))
	  (gtk-entry-select-region entry start-i end-i)
	  (if (not (eq? selected-subtree ex))
	      (begin
		(set! selected-subtree ex)
		(for-each (lambda (x) (x selected-subtree)) rule-filters)))
	  #f)))


    ; add expression to the proof
    (define (add-expression-to-proof expression)
      (set! previous-expressions (cons expression previous-expressions))
      (add-line-to-proof (string-append "     " (unparse expression #f))))

    ; add a rule invocation to the proof
    (define (add-rule-to-proof rule)
      (add-line-to-proof (string-append
			  "<==>        |< "
			  (unparse rule #f)
			  " >|"))
      (set! previous-rules (cons rule previous-rules)))

    ; add string r as the next line in the proof
    (define (add-line-to-proof r)
      (let ((item (gtk-list-item-new))
	    (label (create-math-label r)))
	(gtk-container-add item label)
	(gtk-widget-show-all item)
	(gtk-list-append-item rewriter-list item)))


    (define (set-entry-from-current-expression)
      (set! selected-subtree current-expression)
      (let ((unparse-return
	     (unparse-and-keep-indexes 0 current-expression #f)))
	(set! start-indexes (cadr unparse-return))
	(set! end-indexes (caddr unparse-return))
	(gtk-entry-set-text entry (car unparse-return))
	(for-each (lambda (x) (x selected-subtree)) rule-filters)))

    ; apply the selected rule to the selected subexpression
    (define (applier rule)
      (cond
       ((eq? selected-subtree #f) #f)
       (else
	(let* ((new-expression
		(apply-rule-to-subtree current-expression
				       selected-subtree
				       rule)))
	  (if (eq? new-expression 'fail) 'fail
	      (begin
		(add-rule-to-proof rule)
		(add-expression-to-proof new-expression)
		(set! current-expression new-expression)
		(set-entry-from-current-expression)))))))


    ;; random debugging foo
    (define (debug-clicked)
      (do ((i 0 (+ i 1)))
	  ((= i (length start-indexes)) #t)
	(let ((s (list-ref start-indexes i))
	      (e (list-ref end-indexes i)))
	(display (cadr s))
	(display ", ")
	(display (cadr e))
	(display "    ")
	(display (car s))
	(newline))))


    ;; undo the last rewrite in the proof
    (define (back-clicked)
      (cond
       ((> (length previous-expressions) 1)
	(set! previous-expressions (cdr previous-expressions))
	(set! previous-rules (cdr previous-rules))
	(set! current-expression (car previous-expressions))
	(set-entry-from-current-expression)
	(let ((len (length previous-expressions)))
	  (gtk-list-clear-items rewriter-list
				(- (* len 2) 1)
				(+ (* len 2) 1))))
       (else (display "Can't go back.") (newline))))


    ;; dump the contents of the proof window to stdout
    (define (output-clicked)
      (define (show-one-expression expressions rules)
	(if (eq? expressions '()) #f
	    (begin
	      (display "   ")
	      (display (unparse (car (reverse expressions)) #f))
	      (newline)
	      (show-one-rule (reverse (cdr (reverse expressions))) rules))))
      (define (show-one-rule expressions rules)
	(if (eq? rules '()) #f
	    (begin
	      (display "<==>        |<  ")
	      (display (unparse (car (reverse rules)) #f))
	      (display "  >|")
	      (newline)
	      (show-one-expression
	       expressions
	       (reverse (cdr (reverse rules)))))))
      (show-one-expression previous-expressions previous-rules))

    (define (rabbits-clicked)
      (for-each (lambda (x) (x selected-subtree)) rule-filters))

    (define (quit-clicked)
      (display "quitting...") (newline)
      (gtk-exit))

    ;;;
    ;;;
    ;;;

    (gtk-entry-set-position entry 0)
    (gtk-entry-set-editable entry #f)

    (gtk-container-add (glade-xml-get-widget xml "rules-frame")
		       (create-notebook rules-files givens-files))

    ;; parse the command line
    (set! current-expression (car (parse-gcl cl-expression)))
    (if (expression-contains-rule-variables current-expression)
	(begin
	  (display "don't use variables with leading underscores on ")
	  (display "command line.")
	  (newline)
	  (quit)))
    (set-entry-from-current-expression)

    ;; put the inital line in the proof
    (gtk-list-clear-items rewriter-list 0 -1) ; unneeded?
    (add-expression-to-proof current-expression)

    ;; notice when user clicks on the text-entry
    (gtk-signal-connect* entry "event" entry-event)

    ;; exit gracefully if window is closed.
    (gtk-signal-connect* (glade-xml-get-widget xml "window")
			 "destroy" quit-clicked)


    (gtk-signal-connect* (glade-xml-get-widget xml "quit-menu-item")
			 "activate" quit-clicked)
    (gtk-signal-connect* rabbits-menu-item "activate" rabbits-clicked)
    (gtk-signal-connect* (glade-xml-get-widget xml "debug-menu-item")
			 "activate" debug-clicked)
    (gtk-signal-connect* (glade-xml-get-widget xml "output-menu-item")
			 "activate" output-clicked)
    (gtk-signal-connect* (glade-xml-get-widget xml "back-menu-item")
			 "activate" back-clicked)

    #t))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; start the program
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(define (main args)

  (if (or (null? (cdr args)) (= (string-length (cadr args)) 0))
      (begin (display "Give expression as 1st argument, givens file as 2nd.")
	     (newline)
	     (quit)))

  (glade-init)
  (make-gcl-rewriter (glade-xml-new "gcl-rewriter.glade")
		     (cadr args)
		     (cddr args)) ; list of files containing givens
  (gtk-main))
