;;
;;
;;
;;
;;

;; bigloo module
;;(module dgcl
;;	(main start)
;;	(include "sort.scm")
;;	(include "associators.scm")
;;	(include "lalr.scm")
;;	)


(define binary-infix-operators
  (list '>= '<= '==> '<== '<==> '<=!=> '!=
	'+ '* '/ 'div ; we have no infix "-".  use "x + (-y)" instead.
        'max 'min 'inf 'and ; XXX inf shouldn't be here
	'\/ '/\ '= '> '< 'comma 'dot 'col))
(define unary-prefix-operators
  (list '- '~))
(define quant-operators
  (list 'OR 'AND 'MAX 'MIN 'SUM 'PROD 'N))
(define binary-symmetric-and-associative-operators
  (list '<==> '<=!=> '+ '* 'max 'min '\/ '/\))

;; higher number means it binds more tightly.  make sure this stays
;;  in sync with the grammar.
(define binding-power
  (list '(col 0)
	'(AND 0) '(OR 0) '(MAX 0) '(MIN 0) '(SUM 0) '(PROD 0) '(N 0)
	'(<==> 1) '(<=!=> 1)
	'(==> 2)  '(<== 2)
	'(\/ 3)  '(/\ 3)
	'(= 4)  '(< 4)  '(> 4)  '(<= 4)  '(>= 4)  '(!= 4)
	'(+ 5)
	'(min 6)  '(max 6)  '(* 6)  '(/ 6)  '(div 6)
	'(~ 7)  '(- 7)  '(comma 7)
	'(dot 8)))

;; if this was run with guile-gtk, fall back on the scheme parser.
;;  if this was run from gcl-rewriter, then the yacc/flex generated
;;  parser should already be defined.
(if (not (defined? 'parse-gcl))
    (load "gcl-parser.scm")
    #t)

;; from the little schemer
(define member?
  (lambda (a lat)
    (cond
     ((null? lat) #f)
     (else (or (eq? (car lat) a)
	       (member? a (cdr lat)))))))

(define member-equal?
  (lambda (a lat)
    (cond
     ((null? lat) #f)
     (else (or (equal? (car lat) a)
	       (member-equal? a (cdr lat)))))))


(define (find-parent tree subtree)
  (cond
   ((null? tree) #f)
   ((atom? tree) #f)
   ((member? subtree tree) tree)
   (else
    (let ((r (find-parent (car tree) subtree)))
      (if (not (eq? r #f))
	  r
	  (find-parent (cdr tree) subtree))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; #t if x is an atom, else #f
;; from the little schemer
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define atom?
  (lambda (x)
    (and (not (pair? x)) (not (null? x)))))


;; tree is an ancestor of subtree
(define (subtree-needs-parens? tree subtree)
  (cond
   ((eq? tree subtree) #f) ; avoid putting () around the entire expression
   (else
    (let* ((parent (find-parent tree subtree))
	   (parent-op (if (list? parent) (car parent) #f))
	   (child-op (if (list? subtree) (car subtree) #f))
	   (parent-binding (assv parent-op binding-power))
	   (child-binding (assv child-op binding-power)))
      (cond
       ;; if we don't understand either the parent or child op, put parens.
       ((or (eq? parent-binding #f) (eq? child-binding #f)) #t)
       ;; if the child op has higher binding power than the parent, no parens.
       ((> (cadr child-binding) (cadr parent-binding)) #f)
       ;; if this is part of a comma list, don't unparse to "((q,r),s),t"
       ((and (eq? child-op 'comma) (or (eq? parent-op 'comma)
				       (eq? parent-op 'col))) #f)
       ;; otherwise, be safe and put the parens.
       (else #t))))))


;; call unparse instead of unparse-and-keep-indexes when we don't
;;  care about the indexes.
(define (unparse s show-underscores)
  (car (unparse-and-keep-indexes 0 s show-underscores)))


;; make sure this stays synced with new_infix_exp in gcl-rewriter.c
(define (gcl-symbol->string sym show-underscores)
  (cond
   ((eq? sym 'comma) ",")
   ((eq? sym 'col) ":")
   ((eq? sym 'dot) ".")
   ((and (not show-underscores) (string? sym) (char=? (string-ref sym 0) #\_))
    (substring sym 1 (string-length sym)))
   ((string? sym) sym)
   (else
    (symbol->string sym))))


(define (unparse-and-keep-indexes x expression-tree show-underscores)

  (let ((start-indexes '())
	(end-indexes '()))

    ;; init start-indexes and end-indexes to '('fail 'fail)
    ;; "x" is the starting x index of this expression (unparsed).
    ;; recurse inside here so we have access to the top of the tree
    (define (unparse-and-keep-indexes-rec x tree show-underscores)
      (define (tree->string tree)
	(cond
	 ((string? tree) (gcl-symbol->string tree show-underscores))
	 ((number? tree) (number->string tree))
	 ((symbol? tree) (gcl-symbol->string tree show-underscores))
	 ((boolean? tree) (if tree "true" "false"))
	 ((list? tree)
	  (cond
	   ((null? tree) "")
	   ;; unary prefix
	   ((and (member? (car tree) unary-prefix-operators)
		 (= (length tree) 2))
	    (let* ((need-parens (subtree-needs-parens? expression-tree tree))
		   (paren-width (if need-parens 1 0))
		   (sym-string (gcl-symbol->string (car tree)
						   show-underscores))
		   (sym-length (string-length sym-string))
		   (subtree-string
		    (unparse-and-keep-indexes-rec (+ x paren-width sym-length)
						  (cadr tree)
						  show-underscores)))
	      (string-append (if need-parens "(" "")
			     sym-string
			     subtree-string
			     (if need-parens ")" ""))))
	   ;; binary infix
	   ((and (member? (car tree) binary-infix-operators)
		 (= (length tree) 3))
	    (let* ((need-parens (subtree-needs-parens? expression-tree tree))
		   (paren-width (if need-parens 1 0))
		   (left-subtree-string
		    (unparse-and-keep-indexes-rec (+ x paren-width) (cadr tree)
						  show-underscores))
		   (left-subtree-length (string-length left-subtree-string))
		   ;; put spaces around binary infix unless it is "." or ","
		   ;; put extra spaces on either side of the top level <==>
		   (sym-string
		    (cond
		     ((or (eq? (car tree) 'dot) (eq? (car tree) 'comma))
		      (gcl-symbol->string (car tree) show-underscores))
		     ((and (eq? tree expression-tree) (eq? (car tree) '<==>))
		      (string-append "   "
				     (gcl-symbol->string (car tree)
							 show-underscores)
				     "   "))
		     (else
		      (string-append " "
				     (gcl-symbol->string (car tree)
							 show-underscores)
				     " "))))
		   (sym-length (string-length sym-string))
		   (right-subtree-string (unparse-and-keep-indexes-rec
					  (+ x paren-width
					     left-subtree-length sym-length)
					  (caddr tree)
					  show-underscores)))
	      (string-append (if need-parens "(" "")
			     left-subtree-string
			     sym-string
			     right-subtree-string
			     (if need-parens ")" ""))))
	   ;; quant expressions
	   ((and (member (car tree) quant-operators)
		 (= (length tree) 4))
	    (let* ((sym-string (gcl-symbol->string
				(car tree)
				show-underscores))
		   (sym-length (string-length sym-string))
		   ;;
		   (left-subtree-string (unparse-and-keep-indexes-rec
					 (+ x 2 sym-length) ; 2 is ( and space
					 (cadr tree)
					 show-underscores))
		   (left-subtree-length (string-length left-subtree-string))
		   ;;
		   (mid-subtree-string (unparse-and-keep-indexes-rec
					;; 5 is ( space*3 :
					(+ x 5 sym-length left-subtree-length)
					(caddr tree)
					show-underscores))
		   (mid-subtree-length (string-length mid-subtree-string))
		   ;;
		   (right-subtree-string (unparse-and-keep-indexes-rec
					  ;; 8 is ( space*5 :*2
					  (+ x 8 sym-length left-subtree-length
					     mid-subtree-length)
					  (cadddr tree)
					  show-underscores)))
	      (string-append "("
			     sym-string
			     " "
			     left-subtree-string
			     " : "
			     mid-subtree-string
			     " : "
			     right-subtree-string
			     ")")))
	   (else (display "Unknown operator:")
		 (display (car tree)) (newline))))
	 (else (display "Unknown type:") (display tree) (newline))))

      ;;(display "   ") (display x) (display "   ")
      ;;(display tree) (newline)

      (let* ((tree-string (tree->string tree))
	     (tree-length (string-length tree-string)))
	;;(list-append! start-indexes (list (list tree x)))
	;;(list-append! end-indexes (list (list tree tree-length)))
	(set! start-indexes
	      (cons (list tree x) start-indexes))
	(set! end-indexes
	      (cons (list tree tree-length) end-indexes))
	tree-string))

    ;; make sure we got passed an parsed expression, not a string
    ;; (if (string? expression-tree) (backtrace))

    ;; call the recursive inner-function
    (let ((return-string (unparse-and-keep-indexes-rec x expression-tree
						       show-underscores)))
      (list return-string start-indexes end-indexes))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; index code (don't look)
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (lookup-index tree inds)
  ;;(let ((p (assv tree inds)))
  (let ((p (assq tree inds)))
    (cond
     ((eqv? p #f) 0)
     (else (cadr p)))))

;; (define (lookup-index tree inds)
;;   (define (lookup-index-foo tree inds)
;;     (let ((p (assv tree inds)))
;;       (cond
;;        ((eqv? p #f) 0)
;;        (else (cadr p)))))
;;   (let ((result (lookup-index-foo tree inds)))
;;     (display "lookup-index: ")
;;     (display result)
;;     (display "   ")
;;     (display tree)
;;     (newline)
;;     result))


;; given and index "i" into an unparsed expression, return the
;;  smallest subtree that contains that index.  start-indexes and
;;  end-indexes are produced by unparse-and-keep-indexes.

(define (index->expression expression-tree i start-indexes end-indexes)
  (define (index->expression-rec tree i start-indexes end-indexes)
    ;;(display tree) (newline)
    ;;(display "   ") (display i) (newline)
    (let ((my-start (lookup-index tree start-indexes))
	  (my-width (lookup-index tree end-indexes)))
      (cond
       ((atom? tree) tree)
       ((null? tree) tree)
       ;; unary prefix
       ((and (member? (car tree) unary-prefix-operators)
	     (= (length tree) 2))
	(let ((paren-width
	       (if (subtree-needs-parens? expression-tree tree) 1 0))
	      (s0 (lookup-index (cadr tree) start-indexes))
	      (w0 (lookup-index (cadr tree) end-indexes)))
	  (cond
	   ;; check for clicks on parens or on the neg
	   ((< i s0) tree) ; left paren
	   ((>= i (- (+ my-start my-width) paren-width)) tree) ; right paren
	   ((< i (+ s0 w0 1)) ; i falls in first child
	    (index->expression-rec (cadr tree) i start-indexes end-indexes))
	   (else tree))))
       ;; binary infix
       ((and (member? (car tree) binary-infix-operators)
	     (= (length tree) 3))
	(let ((paren-width
	       (if (subtree-needs-parens? expression-tree tree) 1 0))
	      (s0 (lookup-index (cadr tree) start-indexes))
	      (w0 (lookup-index (cadr tree) end-indexes))
	      (s1 (lookup-index (caddr tree) start-indexes))
	      (w1 (lookup-index (caddr tree) end-indexes)))
	  (cond
	   ((< i s0) tree) ; left paren
	   ((>= i (- (+ my-start my-width) paren-width)) tree) ; right paren
	   ;; not on a paren... try the children
	   ((and (< i (+ s0 w0)) (>= i s0))
	    (index->expression-rec (cadr tree) i start-indexes end-indexes))
	   ((and (< i (+ s1 w1)) (>= i s1))
	    (index->expression-rec (caddr tree) i start-indexes end-indexes))
	   (else tree))))
       ;; quant expressions
       ((and (member (car tree) quant-operators)
	     (= (length tree) 4))
	(let ((s0 (lookup-index (cadr tree) start-indexes))
	      (w0 (lookup-index (cadr tree) end-indexes))
	      (s1 (lookup-index (caddr tree) start-indexes))
	      (w1 (lookup-index (caddr tree) end-indexes))
	      (s2 (lookup-index (cadddr tree) start-indexes))
	      (w2 (lookup-index (cadddr tree) end-indexes)))
	  (cond
	   ((< i s0) tree) ; left paren
	   ((>= i (- (+ my-start my-width) 1)) tree) ; right paren
					; not on a paren... try the children
	   ((and (< i (+ s0 w0)) (>= i s0))
	    (index->expression-rec (cadr tree) i start-indexes end-indexes))
	   ((and (< i (+ s1 w1)) (>= i s1))
	    (index->expression-rec (caddr tree) i start-indexes end-indexes))
	   ((and (< i (+ s2 w2)) (>= i s2))
	    (index->expression-rec (cadddr tree) i start-indexes end-indexes))
	   (else tree))))
       ;;
       (else
	(display "oops -- node has wrong number of children for operator")
	(newline)
	#f))))

  (index->expression-rec expression-tree i start-indexes end-indexes))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; unify code
;;  (from LISP 3rd Edition, P. H. Winston, B. K. P. Horn)
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; (unify '(color _x _y) '(color apple red) '())
;; (unify '(color apple red) '(color _l _m) '())
;; (unify '(color _x _y) '(color _l _m) '())
;; (unify '(color _x _y) '(color _l _m) '())
;; (unify '(color _x _y) '(color _x _y) '())
;; (unify '(_x with (hair blond))
;;        '((patrick is-a person) with (hair blond)) '())
;; (unify '(_x with (hair blond))
;;        '((patrick is-a _y) with (hair blond)) '())
;; (unify '(_x with (hair blond))
;;        '((patrick is-a _x) with (hair blond)) '())

(define (unify p1 p2 bindings)
  (cond
   ((variable-p? p1) (unify-variable p1 p2 bindings))
   ((variable-p? p2) (unify-variable p2 p1 bindings))
   ((and (atom? p1) (atom? p2)) (unify-atoms p1 p2 bindings))
   ((and (list? p1) (list? p2)) (unify-lists p1 p2 bindings))
   (else 'fail)))

(define (unify-atoms p1 p2 bindings)
  (if (equal? p1 p2) bindings 'fail))

(define (unify-lists p1 p2 bindings)
  (cond
   ((and (null? p1) (null? p2)) bindings)
   ((null? p1) 'fail)
   ((null? p2) 'fail)
   (else (let ((result (unify (car p1) (car p2) bindings)))
	   (if (eq? 'fail result)
	       'fail
	       (unify (cdr p1) (cdr p2) result))))))

(define (unify-variable p1 p2 bindings)
  (let ((binding (find-binding p1 bindings)))
    (if binding
	(unify (extract-value binding) p2 bindings)
	(if (insidep p1 p2 bindings)
	    'fail
	    (add-binding p1 p2 bindings)))))

(define (insidep variable expression bindings)
  (if (equal? variable expression)
      #f
      (inside-or-equal-p variable expression bindings)))

(define (inside-or-equal-p variable expression bindings)
  (cond
   ((equal? variable expression) #t)
   ((atom? expression) #f)
   ((null? expression) #f)
   ((eq? '? (car expression))
    (let ((binding (find-binding expression bindings)))
      (if binding
	  (inside-or-equal-p variable (extract-value binding) bindings)
	  #f)))
   (else (or (inside-or-equal-p variable (car expression) bindings)
	     (inside-or-equal-p variable (cdr expression) bindings)))))

;;;;;;;;;;;;;;
;;
;; bindings
;;
;;;;;;;;;;;;;;

(define (add-binding pattern-variable-expression datum bindings)
  (cons (make-binding
	 (extract-variable pattern-variable-expression)
	 datum)
	bindings))

(define (find-binding pattern-variable-expression binding)
  (assoc (extract-variable pattern-variable-expression) binding))

;(define (variable-p? p)
;  (and (list? p)
;       (and (not (null? p))
;	    (eq? '? (car p)))))

(define (variable-p? p)
  (and
   (string? p)
   (> (string-length p) 0)
   (char=? (string-ref p 0) #\_)))

;(define (extract-variable pattern-variable-expression)
;  (cadr pattern-variable-expression))

(define (extract-variable pattern-variable-expression)
  pattern-variable-expression)

(define (make-binding variable datum)
  (list variable datum))

(define (extract-key binding)
  (car binding))

(define (extract-value binding)
  (cadr binding))


;;;;;;;;;;;;;;;;;;;;;;;
;;
;; rule substitution
;;
;;;;;;;;;;;;;;;;;;;;;;;


(define (apply-rule expr rule)
  (cond
   ((not (list? rule)) (display "bad rule") (newline) (list 'fail '()))
   ((null? rule) (display "bad rule") (newline) (list 'fail '()))
   ((eq? (car rule) '<==>)
    ;; attempt to unify expr with the left hand side of the rule
    (let ((bindings (unify expr (cadr rule) '())))
      (if (eq? bindings 'fail)
	  (list 'fail bindings)
	  (list (rewrite (caddr rule) bindings) bindings))))))


(define (apply-rule-to-subtree tree subtree rule)
  (cond
   ((eqv? tree subtree) (car (apply-rule tree rule)))
   ((atom? tree) tree)
   ((null? tree) '())
   (else ; list
    (let ((a (apply-rule-to-subtree (car tree) subtree rule))
	  (d (apply-rule-to-subtree (cdr tree) subtree rule)))
      (cond
       ((eqv? a 'fail) 'fail)
       ((eqv? d 'fail) 'fail)
       (else (cons a d)))))))



(define (rewrite expr bindings)
  (define (copy a)
    (cond
     ((null? a) a)
     ((string? a) (string-copy a))
     ((list? a) (cons (copy (car a)) (copy (cdr a))))
     (else a)))
  (cond
   ((null? expr) '())
   ((variable-p? expr)
    (let ((r (assoc expr bindings)))
      (if (eq? r #f) expr (copy (cadr r)))))
   ((atom? expr) expr)
   ((list? expr) (cons (rewrite (car expr) bindings)
		       (rewrite (cdr expr) bindings)))
   (else 'fail)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; load rules from a file
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (load-rules file-name)

   (display ";;; loading rules from '")
   (display file-name)
   (display "'") (newline)

   (if #t ; make #t to disable the error handler, #f to enable

       (actually-load-rules file-name)

       (begin
 	(let* ((can-load #t)
 	       (handler
 		(lambda (key args . rest)
 		  (set! can-load #f)
 		  key))
 	       (excptn (catch #t
 			    (lambda ()
 			      (actually-load-rules file-name))
 			    handler)))

 	  (cond ((eq? can-load #f)
 		 (display "can't load ")
 		 (display file-name)
 		 (newline)
 		 (display "  exception: ")
 		 (display excptn)
 		 (newline))
 		(else
 		 (set! doc excptn)))))))


(define (actually-load-rules file-name)
  (let ((fp (open-file-for-load file-name)))
    (do ((input-line (read-line fp 'concat) (read-line fp 'concat))
 	 (rules '() rules))
	;;
 	((eof-object? input-line) (reverse rules))
      ;;
      (let ((rule (car (parse-gcl input-line))))
	(set! rules (cons rule rules))))))


(define (open-file-for-load file-name)
  (if (not (file-exists? file-name))
      (begin
 	(display "no such file: ")
 	(display file-name)
 	(newline)
 	(error "can't load.")
 	#f)
      (let* ((fp (false-if-exception (open-file file-name "r"))))
 	(if (eq? fp #f)
 	    (begin
 	      (display "can't open file ")
 	      (display file-name)
 	      (display "for reading")
 	      (newline)
 	      (error "can't load." #f #f)))
 	fp)))



(define (convert-to-symbol-font p)
  ;; returns: (in-symbol-font in-normal-font how-many-used replacement-car)

  ;; list of characters that can be shown in either a normal font or symbol
  (define same-in-both '(#\! #\# #\% #\& #\( #\) #\* #\+ #\, #\- #\. #\/
			 #\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9 #\0
			 #\: #\; #\< #\= #\> #\? #\{ #\| #\} #\~
			 #\space #\newline #\return))

  (define (match? lchr pattern)
    (cond
     ((null? pattern) #t)
     ((null? lchr) #f)
     ((not (eqv? (car lchr) (car pattern))) #f)
     (else
      (match? (cdr lchr) (cdr pattern)))))

  (cond
   ((match? p (string->list "<==>"))  (list #t #f 4 (integer->char 186)))
   ((match? p (string->list "<=!=>")) (list #t #f 5 (integer->char 185)))
   ((match? p (string->list "==>"))   (list #t #f 3 (integer->char 222)))
   ((match? p (string->list "<=="))   (list #t #f 3 (integer->char 220)))
   ((match? p (string->list "~"))     (list #t #f 1 (integer->char 216)))
   ((match? p (string->list "\\/"))   (list #t #f 2 (integer->char 218)))
   ((match? p (string->list "/\\"))   (list #t #f 2 (integer->char 217)))
   ((match? p (string->list "<="))    (list #t #f 2 (integer->char 163)))
   ((match? p (string->list ">="))    (list #t #f 2 (integer->char 179)))
   ((match? p (string->list "AND"))   (list #t #f 3 (integer->char  34)))
   ((match? p (string->list "OR"))    (list #t #f 2 (integer->char  36)))
   ((match? p (string->list "SUM"))   (list #t #f 3 (integer->char 229)))
   ((match? p (string->list "PROD"))  (list #t #f 4 (integer->char 213)))
   ((match? p (string->list "|<"))    (list #t #f 2 (integer->char 225)))
   ((match? p (string->list ">|"))    (list #t #f 2 (integer->char 241)))
   ((match? p (string->list "inf"))   (list #t #f 3 (integer->char 165)))
   ;; notice characters that can be either
   ((and (>= (length p) 1) (member? (car p) same-in-both))
    (list #t #t 1 (car p)))
   (else
    ;; otherwise, use the non-symbol font.
    (list #f #t 1 (car p)))))


;; Return #t if tree contains variables names with a leading underscore,
;;  else #f.  Variables with leading underscores are used in the rules
;;  files to mark variables that can be replaced (bound to parts of an
;;  expression) during unification of rules and expressions.
(define (expression-contains-rule-variables tree)
  (cond
   ((null? tree) #f)
   ((variable-p? tree) #t)
   ((list? tree) (or (expression-contains-rule-variables (car tree))
		     (expression-contains-rule-variables (cdr tree))))
   (else #f)))


;; Make a flat list of variable names from an expression tree
;;  made up of stuff like ('comma ('comma x y) z).
(define (flatten-comma-tree tree)
  (cond
   ((null? tree) '())
   ((string? tree) (list tree))
   ((and (list? tree) (eq? (car tree) 'comma))
    (append (flatten-comma-tree (cadr tree))
	    (flatten-comma-tree (caddr tree))))
   (else
    '())))


;; By convention, the variable z can't be a function of dummie
;;  variables inside of a quantified expression (in a rules file).
(define (z-is-dummy-free? entire-rule bindings)

  ;; convert a list of rule dummies to a list of expression dummies.
  (define (rule-dummies->expression-dummies unbound-dummies)
    (cond
     ((null? unbound-dummies) '())
     ((list? unbound-dummies)
      (append (rule-dummies->expression-dummies (car unbound-dummies))
	      (rule-dummies->expression-dummies (cdr unbound-dummies))))
     (else
      (let ((found-binding (assoc unbound-dummies bindings)))
	(if (eq? found-binding #f)
	    '()
	    (flatten-comma-tree (cadr found-binding)))))))

  (define (is-function-of? tree vars)
    (cond
     ((null? tree) #f)
     ((list? tree) (or (is-function-of? (car tree) vars)
		       (is-function-of? (cdr tree) vars)))
     ((string? tree) (member-equal? tree vars))
     (else #f)))

  (define (check-zs rule-part dummies)
    (cond
     ((null? rule-part) #t)
     ((list? rule-part)
      (if (member? (car rule-part) quant-operators)
	  (let ((new-dummies (append dummies
				     (flatten-comma-tree (cadr rule-part)))))
	    (and (check-zs (caddr rule-part) new-dummies)
		 (check-zs (cadddr rule-part) new-dummies)))
	  (and (check-zs (car rule-part) dummies)
	       (check-zs (cdr rule-part) dummies))))
     ((and (string? rule-part) (string=? rule-part "_z"))
      (let ((expression-dummies (rule-dummies->expression-dummies dummies))
	    (expression-z (assoc "_z" bindings)))
	(if (not (eq? expression-z #f))
	    (not (is-function-of? (cadr expression-z) expression-dummies))
	    #t)))
     (else #t)))

  (check-zs (cadr entire-rule) '()))





;; display all trees with unique left-hand sides
;; that are equivalent to this one (due to symmetry and associativity).
;; for a good time:
;;  (permute-expression (car (parse-gcl "_p /\ _q<==>(_p<==>_q<==>_p \/ _q)")))

(define (permute-expression original-tree)

  (let ((expressions '()))

    (define (bindings-are-var-to-var bindings)
      ;; return #t if unify succeeded and matched all leaves
      ;; in the rule to leaves in the expression
      ;;(if (and (not (null? bindings)) (list? bindings))
      ;;  (begin (display "    ") (display (car bindings)) (newline)))
      (cond
       ((eq? bindings 'fail) #f)
       ((null? bindings) #t)
       ((atom? (car (cdr (car bindings))))
	(bindings-are-var-to-var (cdr bindings)))
       (else #f)))

    (define left-side-unique?
      (lambda (a lat)
	(cond
	 ((null? lat) #t)
	 ;; only consider an expression unique if its lefthand side
	 ;;  is unlike any of the other's lefthand sides.
	 ((bindings-are-var-to-var (unify (cadr (car lat)) (cadr a) '())) #f)
	 (else (left-side-unique? a (cdr lat))))))

    (define (rule-introduces-new-variables? tree)
      (define (find-vars sub-tree)
	(cond
	 ((null? sub-tree) '())
	 ((and (string? sub-tree) (char=? (string-ref sub-tree 0) #\_))
	  (list sub-tree))
	 ((not (list? sub-tree)) '())
	 (else
	  (let ((car-vars (find-vars (car sub-tree)))
		(cdr-vars (find-vars (cdr sub-tree))))
	    (for-each (lambda (x)
			(if (not (member-equal? x car-vars))
			    (set! car-vars (cons x car-vars))))
		      cdr-vars)
	    car-vars))))
      (let ((left-vars (find-vars (cadr tree)))
	    (right-vars (find-vars (caddr tree)))
	    (new-vars-found #f))
	(for-each (lambda (x)
		    (if (not (member-equal? x left-vars))
			(set! new-vars-found #t)))
		  right-vars)
	new-vars-found))


    (define (add-unique-expression tree)
      (if (left-side-unique? tree expressions)
	  (begin
	    ;; don't show the ones that will cause new variables
	    ;;  to enter the expression
	    (if (not (rule-introduces-new-variables? tree))
		(begin
		  (display (unparse tree #t))
		  (newline)))
	    ;; get rid of _ in var names and add to list
	    (set! expressions (cons (car (parse-gcl (unparse tree #f)))
				    expressions))
	    #t)
	  #f))

    ;; copy tree, replace subtree with replacement
    (define (splice-tree tree subtree replacement)
      (cond
       ((eqv? tree subtree) replacement)
       ((atom? tree) tree)
       ((null? tree) '())
       (else ; list
	(let ((a (splice-tree (car tree) subtree replacement))
	      (d (splice-tree (cdr tree) subtree replacement)))
	  (cons a d)))))

    (define (permute-subtree tree subtree)
      (add-unique-expression tree)
      (if (list? subtree)
	  (begin
	    ; (op0 (op1 w x) (op2 y z))
	    ; (op0 (op1 w x))
	    ; (op0 x)
	    (permute-subtree tree (cadr subtree))
	    (if (not (null? (cddr subtree)))
		(permute-subtree tree (caddr subtree)))
	    ;; if the operator is symmetric, swap the arguments
	    (if (member? (car subtree)
			 binary-symmetric-and-associative-operators)
		(let* ((new-subtree (list (car subtree)
					  (caddr subtree)
					  (cadr subtree)))
		       (new-tree (splice-tree tree subtree new-subtree)))
		  (if (add-unique-expression new-tree)
		      (permute-subtree new-tree new-tree))))
	    ;; if the operator is associative, trade like this:
	    ;;  (x+y)+z  becomes  (x+z)+y
 	    (if (and (member? (car subtree)
			      binary-symmetric-and-associative-operators)
		     (list? (cadr subtree))
 		     (eqv? (car subtree) (car (cadr subtree))))
 		(let* ((new-subtree (list (car subtree)
 					  (list (car subtree)
 						(cadr (cadr subtree))
 						(caddr subtree))
 					  (caddr (cadr subtree))))
 		       (new-tree (splice-tree tree subtree new-subtree)))
 		  (if (add-unique-expression new-tree)
 		      (permute-subtree new-tree new-tree))))
	    )))

    (permute-subtree original-tree original-tree)

    ;;(display "-----------------------------------------") (newline)
    ;;(for-each (lambda (x) (display (unparse x #t)) (newline)) expressions)
    ))
